//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Base class for creating random knots.

#include <GraphicsDefs.h>
#include <Point.h>
#include <Rect.h>
#include <View.h>
#include <Window.h>

#include "KnotSection.h"
#include "KnotSectionCache.h"
#include "RandomColor.h"
#include "RandomKnot.h"
#include "RandomNumbers.h"

// Generate an array of random direction values.
static direction *
randomDirectionArray(int width, int height, int modulus)
{
	direction *pArray = new direction[width * height];
	for (int i = 0; i < height; ++i) {
		for (int j = 0; j < width; ++j) {
			// Prefer diagonal sections resp. corners.
			switch (randomInteger(modulus)) {
			
			case 0:
				pArray[i*width + j] = H;
				break;
			
			case 1:
				pArray[i*width + j] = V;
				break;
			
			default:
				pArray[i*width + j] = D;
				break;
			}
		}
	}
	
	return pArray;
}

RandomKnot::RandomKnot(
	BView *pView, int maxSections, int xRepeat, int yRepeat)
:	mpView(pView),
	mWidth(3 + randomInteger(maxSections - 3)),
	mHeight(3 + randomInteger(maxSections - 3)),
	mpSectionTypes(0),
	mpSectionCorners(0),
	mpCache(0),
	mColor(RandomColor(0.2, 0.8, 0.5, 1.0))
{
	if (mpView->LockLooper()) {
		// Calculate dimensions of the knot.
		BRect bounds = mpView->Bounds();
		int sectionWidth = bounds.IntegerWidth() / mWidth / xRepeat / 2;
		int sectionHeight = bounds.IntegerHeight() / mHeight / yRepeat / 2;
		float edgeLength =
			(float)((sectionWidth < sectionHeight)
						? (1.0 + randomDouble()) * sectionWidth
						: (1.0 + randomDouble()) * sectionHeight);
		
		if (edgeLength >= 10.0) {
			// Determine knot style.
			bool slim = randomDouble() < 0.6;
			double outline = slim ? 0.25 : 0.5;
			double fill = outline * 0.8;
			bool hollow = randomDouble() < 0.5
						  && (edgeLength * (outline - fill)) > 3;

			// Create the section cache.
			mpCache = new KnotSectionCache(edgeLength, outline, fill, hollow);

			// Initialize the knot data.  Derived classes have to establish
			// the proper boundary conditions.
			mpSectionTypes = randomDirectionArray(mWidth, mHeight, 17);
			mpSectionCorners = randomDirectionArray(mWidth+1, mHeight+1, 5);
		}
		
		mpView->UnlockLooper();
	}
}

RandomKnot::~RandomKnot(void)
{
	delete mpCache;
	mpCache = 0;
	delete [] mpSectionTypes;
	mpSectionTypes = 0;
	delete [] mpSectionCorners;
	mpSectionCorners = 0;
}

BPoint
RandomKnot::LocateKnot(void)
{
	if (!IsValid()) {
		return B_ORIGIN;
	}

	// Find a random position where the knot will be visible.
	BRect bounds = mpView->Bounds();
	const int edgeLength = (int)(mpCache->EdgeLength());

	BPoint where(
		(float)randomInteger(bounds.IntegerWidth() - mWidth*edgeLength),
		(float)randomInteger(bounds.IntegerHeight() - mHeight*edgeLength));
	
	return where;
}

void
RandomKnot::DrawAt(BPoint where)
{
	if (!IsValid()) {
		return;
	}
	
	// Draw all the sections.
	mpView->SetHighColor(mColor);
	mpView->SetDrawingMode(B_OP_OVER);
	const float offset = mpCache->EdgeLength();
	for (int i = 0; i < mHeight; ++i) {
		for (int j = 0; j < mWidth; ++j) {
			KnotSection section(mpSectionTypes[i*mWidth + j],
									  mpSectionCorners[i*(mWidth+1) + j],
									  mpSectionCorners[i*(mWidth+1) + j + 1],
									  mpSectionCorners[(i+1)*(mWidth+1) + j],
									  mpSectionCorners[(i+1)*(mWidth+1) + j + 1]);
			mpView->DrawBitmapAsync(mpCache->GetKnotSection(section),
											where + BPoint(j * offset, i * offset));
		}
	}
}
