//		Copyright © 1999–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Screen saver class.

#include <cstring>

#include <Bitmap.h>
#include <DataIO.h>
#include <File.h>
#include <GraphicsDefs.h>
#include <image.h>
#include <Message.h>
#include <OS.h>
#include <Point.h>
#include <Rect.h>
#include <Resources.h>
#include <ScreenSaver.h>
#include <TextView.h>
#include <TranslationUtils.h>
#include <View.h>

#include "ClosedKnot.h"
#include "HorizontalKnot.h"
#include "RandomNumbers.h"
#include "Samhain.h"
#include "SamhainScreenSaver.h"
#include "VerticalKnot.h"

const rgb_color kBlack = { 0, 0, 0, 255 };

// Screen saver interface.
BScreenSaver *
instantiate_screen_saver(BMessage *pMsg, image_id id)
{
	return new SamhainScreenSaver(pMsg, id);
}

SamhainScreenSaver::SamhainScreenSaver(BMessage *pMsg, image_id id)
:	inherited(pMsg, id),
	mPreviewing(false),
	mDelay(0),
	mKnotsDrawn(0),
	mLogoCounter(0),
	mpLogo(0)
{
	srand((long)system_time());

	// Try to load a logo bitmap from the add-on's resources.
	image_info info;
	if (get_image_info(id, &info) == B_OK) {
		BFile file(info.name, B_READ_ONLY);
		BResources resources(&file);
		size_t size;
		const void *pData =
			resources.LoadResource(B_RAW_TYPE, "BeOS logo data", &size);
		if (pData) {
			BMemoryIO stream(pData, size);
			mpLogo = BTranslationUtils::GetBitmap(&stream);
		}
		
		// ### For now, the alpha channel is stored in a seaparate bitmap.
		if (mpLogo && mpLogo->ColorSpace() == B_RGB32) {
			const void *pMask =
				resources.LoadResource(B_RAW_TYPE, "BeOS logo mask", &size);
			if (pMask) {
				BMemoryIO stream(pMask, size);
				BBitmap *pMaskBitmap = BTranslationUtils::GetBitmap(&stream);
				if (pMaskBitmap && pMaskBitmap->ColorSpace() == B_RGB32) {
					uint8 *pDataBits = (uint8 *)(mpLogo->Bits());
					uint8 *pMaskBits = (uint8 *)(pMaskBitmap->Bits());
					for (size_t n = mpLogo->BitsLength(); n >= sizeof(uint32); n -= sizeof(uint32)) {
						pDataBits[3] = pMaskBits[0];
						pDataBits += sizeof(uint32);
						pMaskBits += sizeof(uint32);
					}
				}
				delete pMaskBitmap;
			}
		}
	}
	
	if (mpLogo) {
		mLogoBounds = mpLogo->Bounds();
	}
}

SamhainScreenSaver::~SamhainScreenSaver(void)
{
	delete mpLogo;
	mpLogo = 0;
}

void
SamhainScreenSaver::Draw(BView *pView, int32 frame)
{
	if (frame > 0 && --mDelay > 0) {
		return;
	}
	
	if (pView->LockLooper()) {
		const BRect bounds = pView->Bounds();
		
		if (frame == 0
			 || ++mKnotsDrawn > 4
			 || randomDouble() > 0.7) {
			pView->SetHighColor(kBlack);
			pView->SetDrawingMode(B_OP_COPY);
			pView->FillRect(bounds);
			mKnotsDrawn = 0;
		}

		if (!mPreviewing && mpLogo && (mLogoCounter += randomInteger(3)) > 50) {
			// Let's hear it for the BeOS!
			const int xPos =
				randomInteger(bounds.IntegerWidth() - mLogoBounds.IntegerWidth());
			const int yPos =
				randomInteger(bounds.IntegerHeight()- mLogoBounds.IntegerHeight());
			pView->SetDrawingMode(B_OP_ALPHA);
			pView->SetBlendingMode(B_PIXEL_ALPHA, B_ALPHA_OVERLAY);
			pView->DrawBitmap(mpLogo, BPoint(xPos, yPos));
			mLogoCounter = 0;
		
		} else {
			// Draw any old knot.
			int type = randomInteger(5);
			if (type > 1 || mPreviewing) {
				bool drawn;
				do {
					ClosedKnot knot(pView);
					drawn = knot.Draw();
				} while (!(mPreviewing || drawn));

			} else if (type == 0) {
				bool drawn;
				do {
					HorizontalKnot knot(pView);
					drawn = knot.Draw();
				} while (!drawn);

			} else {
				bool drawn;
				do {
					VerticalKnot knot(pView);
					drawn = knot.Draw();
				} while (!drawn);
			}
		}
		pView->UnlockLooper();
	}

	mDelay = mPreviewing ? 1 : 3 + randomInteger(4);
}

void
SamhainScreenSaver::StartConfig(BView *pConfigView)
{
	if (pConfigView->LockLooper()) {
		BRect frame = pConfigView->Bounds();
		frame.top += 10.0;
		frame.left += 10.0;
		BRect textArea = frame;
		textArea.OffsetTo(B_ORIGIN);
		BTextView *pCaption =
			new BTextView(frame, B_EMPTY_STRING, textArea,
								B_FOLLOW_ALL, B_WILL_DRAW);
		pCaption->SetText("Samhain " SAMHAIN_VERSION "\n"
								"\n"
								"\tCeltic knotwork à la Andy Sloss.\n" 
								"\t© 1997–2000  Jens Kilian <jjk@acm.org>\n"
								"\n"
								"Samhain comes with ABSOLUTELY NO WARRANTY; "
								"it is free software, and you are welcome "
								"to redistribute it under certain conditions.  "
								"See the GNU General Public License for details.");
		pConfigView->AddChild(pCaption);
		pCaption->SetViewColor(pConfigView->ViewColor());
		pCaption->SetStylable(true);
		pCaption->SetFontAndColor(be_plain_font);
		const char *pText = pCaption->Text();
		pCaption->SetFontAndColor(0, strchr(pText, '\n') - pText, be_bold_font);
		pCaption->MakeEditable(FALSE);
		pCaption->MakeSelectable(FALSE);
		
		pConfigView->UnlockLooper();
	}
}

status_t
SamhainScreenSaver::StartSaver(BView *, bool preview)
{
	SetTickSize(1000000);	// 1 second

	mPreviewing = preview;
	mDelay = mKnotsDrawn = mLogoCounter = 0;

	return B_OK;
}
